using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;
using System.Threading;
using System.Diagnostics;

namespace Danga 
{
    public class MonoBal
    {  
	int inChunks = 0;   // chunks in queues from clients to proxies (POSTs)
	int outChunks = 0;  // chunks in queues from proxies to clients (responses)

	struct Chunk {
	    public int    Size;
	    public byte[] Buf;

	    public Chunk (int size, byte[] buf) {
		Size = size;
		Buf = buf;
	    }
	}

	class ClientHandler
	{
	    const int       BufferSize = 4096;
	    byte[]          buffer;

	    Socket          cSock;
	    UsefulStream    cUS;    // for reading headers/POST data
	    NetworkStream   cNS;    // stream for writing back responses
	    

	    Queue           inQ;
	    Queue           outQ;

	    // Client State
	    string clientMethod, clientURI, clientHTTPVer;
	    HTTPHeaders clientHeaders;

	    private int  clientRead = 0;  // bytes read from client, zeroed at each reading stage
	    private AsyncCallback   clientReadCallback;
	    private byte[]  requestBytes;

	    // Proxy State
	    NetworkStream pNS;
	    Socket pSock;

	    // constructor
	    public ClientHandler(Socket socketForClient)
	    {
		cSock = socketForClient;
		cUS = new UsefulStream(socketForClient); // for reading
		cNS = new NetworkStream(cSock);
		
		inQ = Queue.Synchronized(new Queue());
		outQ = Queue.Synchronized(new Queue());
	    }

	    // entry point for a new client connection.  just tells the usefulstream
	    // to begin a line read, which is the Request Line.
	    public void StartRead ()
	    {
		cUS.BeginGetLine(new AsyncCallback(this.OnGetRequestLine), 0, null);
	    }

	    // called for the first line of an HTTP request.
	    // doesn't allow proxy requests.  just /abs_path or *.  HTTP version is optional and
	    // otherwise assumes old 0.9-style.  starts subsequent header reads.
	    private void OnGetRequestLine (IAsyncResult ar) {
		string s = cUS.EndGetLine();
		Regex reVerb = new Regex(@"^(?<method>\w+) (?<uri>(?:\*|(?:/\S*?)))(?: HTTP/(?<ver>\d+\.\d+))?\r?\n?$");
		Match m = reVerb.Match(s);
		if (! m.Success) {
		    Close();
		    return;
		}

		clientMethod = m.Groups["method"].ToString();
		clientURI = m.Groups["uri"].ToString();
		clientHTTPVer = m.Groups["ver"].ToString();
		clientHeaders = new HTTPHeaders ();
		clientHeaders.RequestLine = clientMethod + " " + clientURI + " HTTP/1.0\r\n";

		cUS.BeginGetLine(new AsyncCallback(this.OnGetHeader), 0, null);
	    }

	    // callback for each HTTP header line 
	    private void OnGetHeader (IAsyncResult ar) {
		String s = cUS.EndGetLine();
		if (s == null) { 
		    Trace.WriteLine("s == null.  Closing connection!");
		    Close(); 
		    return; 
		}

		//Trace.WriteLine("got header: {0}", s);

		if (! clientHeaders.ParseNewHeader(s)) {
		    Trace.WriteLine("ParseNewHeader fails. Closing connection!");
		    Close();
		    return;
		}

		//Trace.WriteLine("Headers: [{0}]", clientHeaders);
		
		if (clientHeaders.Complete) {
		    string scl = clientHeaders.GetHeader("Content-Length");
		    int cl = 0;
		    if (scl != null) {
			try {
			    cl = int.Parse(scl);
			} catch {
			    Close();
			    return;
			}
		    }
		    //Trace.WriteLine("content length: {0}", cl);

		    clientHeaders.SetHeader("X-Forwarded-For", 
					    ((IPEndPoint)cSock.RemoteEndPoint).Address.ToString());
		    clientHeaders.SetHeader("Connection", "close");

		    Trace.WriteLine("Final Headers: [" + clientHeaders + "]");

		    ProxyConnect();
		    return;
		}

		// start read of a new header line
		cUS.BeginGetLine(new AsyncCallback(this.OnGetHeader), 0, null);
	    }


		

	    // start a connect to some proxy
	    private void ProxyConnect () {
		//FIXME: pick one from the proxy selection manager
		IPAddress kenny = IPAddress.Parse("10.1.0.2");
		IPEndPoint pxEP = new IPEndPoint(kenny, 80);
		pSock = new Socket(pxEP.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
		pSock.BeginConnect(pxEP,
				   new AsyncCallback(OnProxyConnect), pSock);
	    }

	    private void OnProxyConnect ( IAsyncResult ar ) {
		if (! pSock.Connected) {
		    Trace.WriteLine("Proxy connect errro!");
		    Close();
		    return;
		}
		pNS = new NetworkStream(pSock);

		// send the request
		ProxyWrite(clientHeaders.GetBytes());

		// then any extra buffer leftovers from UsefulStream
		// as it switches from line-reading to byte reading.
		// (lame, yeah)
		ProxyWrite(cUS.GetExtra());

		ClientReadMore();  // and read from the client too
	    }

	    private void ProxyWrite (byte[] buf) {
		if (buf != null)
		    ProxyWrite(buf, buf.Length);
	    }

	    private void ProxyWrite (byte[] buf, int len) {
		if (buf == null || len == 0) return;
		Trace.WriteLine("proxywrite, len=" + len);
		inQ.Enqueue(new Chunk(len, buf));
		ProxyPush();
	    }

	    private void ProxyReadMore () {
		byte[] buf = new byte[4096];
		pNS.BeginRead(buf, 0, 4096, new AsyncCallback(this.onProxyRead), buf);
	    }

	    private void ClientReadMore () {
		byte[] buf = new byte[4096];
		if (cUS == null) return;
		NetworkStream cInNS = cUS.Stream;
		if (cInNS == null) return;
		cInNS.BeginRead(buf, 0, 4096, new AsyncCallback(this.onClientRead), buf);
	    }

	    private void onProxyRead (IAsyncResult ar) {
		int bytesRead = pNS.EndRead(ar);
		byte[] buf = (byte[]) ar.AsyncState;

		Trace.WriteLine("onProxyRead = bytes: " + bytesRead);

		//Trace.WriteLine("proxy read = {0}", bytesRead);
		if (bytesRead == 0) {
		    ProxyClose();
		    // signal the end of the stream with a 0/null chunk
		    outQ.Enqueue(new Chunk(0, null));
		} else {
		    outQ.Enqueue(new Chunk(bytesRead, buf));
		}
		ClientPush();

		if (bytesRead > 0) {  // TODO: and not too much in the queue
		    ProxyReadMore();
		}
	    }

	    private void onClientRead (IAsyncResult ar) {
		Trace.WriteLine("on client read 1");
		if (cUS == null) return;
		NetworkStream ns = cUS.Stream;
		if (ns == null) return;
		Trace.WriteLine("onClientRead 2");
		int bytesRead = 0;
		try {
		    bytesRead = ns.EndRead(ar);
		    Trace.WriteLine("onClientRead 3");
		} catch {
		    Trace.WriteLine("Error reading from client");
		    Close();
		    return;
		}
		Trace.WriteLine("onClientRead 4");
		byte[] buf = (byte[]) ar.AsyncState;
		Trace.WriteLine("onClientRead 5");

		Trace.WriteLine("bytes from client: " + bytesRead);
		inQ.Enqueue(new Chunk(bytesRead, buf));
		ProxyPush();

		if (bytesRead > 0) {  // TODO: and not too much in the queue
		    ClientReadMore();
		}
	    }

	    
	    int clientWriting = 0;
	    // called to instigate writes to client, if isn't one going on already
	    private void ClientPush () {
		if (Interlocked.CompareExchange(ref clientWriting, 1, 0) > 0)
		    return;

		//Trace.WriteLine("clientWriting!");

		if (outQ.Count == 0) {
		    clientWriting = 0;
		    //Trace.WriteLine("outQ empty");
		    return;
		}

		Chunk c = (Chunk) outQ.Dequeue();
		//Trace.WriteLine("client push = {0}", c.Size);

		if (c.Size != 0) {
		    try {
			Trace.WriteLine("beginning write to client of bytes: " + c.Size);
			cNS.BeginWrite(c.Buf, 0, c.Size, 
				       new AsyncCallback(this.onClientWrite), null);
		    } catch {
			Trace.WriteLine("exception after cNS.BeginWrite");
			Close();
			return;
		    }
		} else {
		    // done!
		    Trace.WriteLine("Done to client!");
		    Close();
		    return;
		}
	    }

	    int proxyWriting = 0;
	    private void ProxyPush () {
		if (Interlocked.CompareExchange(ref proxyWriting, 1, 0) > 0)
		    return;

		if (inQ.Count == 0) {
		    proxyWriting = 0;
		    return;
		}

		Chunk c = (Chunk) inQ.Dequeue();

		if (c.Size != 0) {
		    try {
			Trace.WriteLine("writing to proxy: bytes=" + c.Size);
			pNS.BeginWrite(c.Buf, 0, c.Size, 
				       new AsyncCallback(this.onProxyWrite), null);
		    } catch {
			Trace.WriteLine("exception after cPS.BeginWrite");
			Close();
			return;
		    }
		} else {
		    Close();
		    return;
		}
	    }

	    private void onClientWrite (IAsyncResult ar) {
		Trace.WriteLine("ending client write");
		try {
		    cNS.EndWrite(ar);
		} catch {
		    Trace.WriteLine("Client died or something");
		    Close();
		    return;
		}
		clientWriting = 0;
		ClientPush();
	    }

	    private bool sentProxyRequest = false;
	    private void onProxyWrite (IAsyncResult ar) {
		try {
		    pNS.EndWrite(ar);
		} catch {
		    Trace.WriteLine("Proxy died or something, shutting down");
		    Close();
		    return;
		}
		Trace.WriteLine("did proxy write");
		proxyWriting = 0;

		// if this is the first write to the proxy, it's the request, so
		// now we can potentially start reading.
		if (! sentProxyRequest) {
		    sentProxyRequest = true;
		    ProxyReadMore();
		}
		ProxyPush();
	    }

	    private void ProxyClose () {
		if (pNS == null || pSock == null) return;
		pNS.Close();
		if (pSock.Connected) {
		    pSock.Shutdown(SocketShutdown.Both);
		    pSock.Close();
		}
		pSock = null;
		pNS = null;
	    }

	    private void ClientClose () {
		if (cNS != null) {
		    try {
			cNS.Close();
		    } catch {
			Trace.WriteLine("exception closing cNS?");
		    }
		}
		if (cUS != null) {
		    try {
			cUS.Close();
		    } catch {
			Trace.WriteLine("exception closing cNS?");
		    }
		}
	    }

	    private void Close () {
		ClientClose();
		ProxyClose();
	    }
	}


	private void Run( )
	{
	    IPAddress localAddr = IPAddress.Parse("0.0.0.0");
	    TcpListener tcpListener = new TcpListener(localAddr, 65000);
	    tcpListener.Start( );       

	    for (;;) {
		Socket socketForClient = 
		    tcpListener.AcceptSocket();
		if (socketForClient.Connected)
		    {
			//Trace.WriteLine("Client connected");
			ClientHandler handler = 
			    new ClientHandler(socketForClient);
			handler.StartRead( );
		    }
	    }
	}

	public static void Main ()
	{
	    MonoBal app = new MonoBal();
	    app.Run();
	}

    } // public class MonoBal

} // namespace Danga


       
